// Copyright David Abrahams 2004. Distributed under the Boost
// Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
#ifndef SFINAE_DWA2004723_HPP
# define SFINAE_DWA2004723_HPP

# include <boost/python/detail/prefix.hpp>

# if defined(BOOST_NO_SFINAE) && !defined(BOOST_MSVC)
#  define BOOST_PYTHON_NO_SFINAE
# endif

#endif // SFINAE_DWA2004723_HPP
